# pkg:gem/syntax_tree@SYNTAX_TREE_VERSION#lib/syntax_tree.rb:39
def foo
end

# pkg:gem/open3#open3.rb:1
def bar
end

# pkg:gem/syntax_tree#lib/syntax_tree.rb:39
def baz
end

# pkg:gem/syntax_tree#lib/syntax_tree.rb:1
class Foo
end

# pkg:gem/syntax_tree#lib/syntax_tree.rb:2
class Foo::Bar
end

# pkg:gem/syntax_tree#lib/syntax_tree.rb:3
module Foo
end

# pkg:gem/syntax_tree#lib/syntax_tree.rb:4
FOO = 1

# pkg:gem/syntax_tree#lib/syntax_tree.rb:5
FOO::BAR = 1

# pkg:gem/deleted#lib/foo.rb.rb:1
def baz
end
