attr_reader :a, "b", some_variable
attr_writer :c, "d", some_variable
attr_accessor :e, "f", some_variable

self.attr_reader :a, "b", some_variable
self.attr_writer :c, "d", some_variable
self.attr_accessor :e, "f", some_variable
