# This is the documentation for Foo::Bar
class Foo::Bar
end

# This is more documentation for Foo::Bar
class Foo::Bar
end

class Foo
  # This is even more documentation for Foo::Bar
  class Bar
  end
end
